# $Id: SyncServer.pm 364 2013-07-02 13:00:35Z phil $
#
# author: Philippe Kueck <projects@unixadm.org>
#

package SyncServer;

use strict;
use warnings;

use FCGI;
use IO::Handle;
use MIME::Base64;
use JSON;
use Time::HiRes qw(time);
use SyncServer::DB;
use SyncServer::Config;

sub new {
	return unless SyncServer::Config::load;
	my $self = {
		'in' => new IO::Handle,
		'out' => new IO::Handle,
		'err' => new IO::Handle,
		'db' => new SyncServer::DB,
		'env' => undef
	};

	$self->{'req'} = FCGI::Request(
		$self->{'in'},
		$self->{'out'},
		$self->{'err'},
		\%{$self->{'env'}}
	);

	if ($config->{'ldap'}) {
		use SyncServer::LDAP;
		$self->{'ldap'} = new SyncServer::LDAP;
		return unless $self->{'ldap'}
	}

	return unless $self->{'req'};
	return unless $self->{'db'};

	return bless $self, $_[0]
}

sub init {
	delete $_[0]->{'data'};
	foreach (keys %{$_[0]->{'qpars'}}) {delete $_[0]->{'qpars'}->{$_}}
	unless ($_[0]->{'db'}->reconnect) {
		printf "*** Database error: %s\n", $_[0]->{'db'}->error
	}
}

sub header {
	my ($self, $code, $msg) = @_;
	my $ret = {
		200 => 'OK',
		400 => 'Bad Request',
		401 => 'Not Authorized',
		403 => 'Forbidden',
		404 => 'Not Found',
		503 => 'Error'
	};
	$code = 200 unless $code;
	if (!exists $ret->{$code}) {
		if (defined $msg) {$ret->{$code} = $msg}
		else {$ret->{$code} = "Unknown"}
	}
	
	my $queue;
	$queue .= sprintf "Status: %d %s\r\n", $code, $ret->{$code};
	$queue .= sprintf "Content-Type: %s\r\n",
		$self->{'content-type'}?$self->{'content-type'}:"text/plain";
	$queue .= sprintf "Retry-After: 3600\r\n" if $code == 503;
	$queue .= "Strict-Transport-Security: max-age=2419200\r\n";
	$queue .= sprintf "X-Weave-Timestamp: %.2f\r\n", time;
	$queue .= sprintf "WWW-Authenticate: basic realm=\"syncserver\"\r\n"
		if $code == 401;
	$queue .= sprintf "X-Weave-Alert: %s\r\n", $msg if defined $msg;
	$queue .= sprintf "Content-Length: %d\r\n", $self->{'content-length'}
		if defined $self->{'content-length'};
	$queue .= "\r\n";

	printf {$self->{'out'}} $queue;
	delete $self->{'content-length'};
	delete $self->{'content-type'}
}

sub send {
	my ($self, $msg, @args) = @_;
	printf {$self->{'out'}} $msg, @args
}

sub setcontentlength {
	$_[0]->{'content-length'} = $_[1]
}

sub setcontenttype {
	$_[0]->{'content-type'} = $_[1]
}

sub data {
	return unless $_[0]->{'env'}->{'CONTENT_LENGTH'} &&
		$_[0]->{'env'}->{'CONTENT_LENGTH'} > 0;
	my ($data, $tmp);
	for (my $i = 0; $i < $_[0]->{'env'}->{'CONTENT_LENGTH'}; $i++) {
		$tmp = getc($_[0]->{'in'});
		return if $tmp eq '';
		$data .= $tmp
	}
	if ($data =~ m/^[\[\{].*[\}\]]$/) {
		eval {$tmp = decode_json($data)};
		return if $@;
		$_[0]->{'data'} = $tmp;
		return $tmp
	}
	$_[0]->{'data'} = $data;
	return $data
}

sub dumpenv {
	foreach (keys %{$_[0]->{'env'}}) {
		printf "||| \$env{'%s'} = %s\n", $_, $_[0]->{'env'}->{$_}
	}
	foreach (keys %{$_[0]->{'qpars'}}) {
		printf "||| \$qpars{'%s'} = %s\n", $_, $_[0]->{'qpars'}->{$_}
	}
}

sub env {
	$_[0]->{'env'}->{$_[1]} = $_[2] if defined $_[2];
	return $_[0]->{'env'}->{$_[1]}
}

sub qpars {
	$_[0]->{'qpars'}->{$_[1]} = $_[2] if defined $_[2];
	return $_[0]->{'qpars'}->{$_[1]}
}

sub requestIs {return ($_[0]->{'env'}->{'REQUEST_METHOD'} eq $_[1])?1:0}

sub Accept {return $_[0]->{'req'}->Accept}

sub Flush {return $_[0]->{'req'}->Flush}

sub GetEnvironment {return $_[0]->{'req'}->GetEnvironment}

sub auth {
	return 0 unless $_[0]->{'env'}->{'SYNC_USER'};
	return 0 unless $_[0]->{'env'}->{'AUTHORIZATION'};
	($_[0]->{'env'}->{'REMOTE_USER'}, $_[0]->{'env'}->{'REMOTE_PASSWD'}) =
		split ':', decode_base64((split /\s+/, $_[0]->{'env'}->{'AUTHORIZATION'})[1]) if
		$_[0]->{'env'}->{'AUTHORIZATION'};
	return 0 unless $_[0]->{'env'}->{'REMOTE_PASSWD'};
	return 0 unless $_[0]->{'env'}->{'REMOTE_USER'} eq $_[0]->{'env'}->{'SYNC_USER'};
	return 0 if $config->{'ldap'} && !$_[0]->{'ldap'}->getuser($_[0]->{'env'}->{'SYNC_USER'});
	return $_[0]->{$config->{'ldap'}?'ldap':'db'}->auth(
		$_[0]->{'env'}->{'SYNC_USER'}, $_[0]->{'env'}->{'REMOTE_PASSWD'}
	)
}


1;

__END__

=head1 NAME

SyncServer - The SyncServer core

=head1 SYNOPSIS

 use SyncServer;

=head1 DESCRIPTION

This is the SyncServer core. It is mainly used by L<SyncServer::Storage> and L<SyncServer::User>.

=head1 DEPENDENCIES

=over 8

=item FCGI

=item IO::Handle

=item MIME::Base64

=item JSON

=item Time::HiRes

=item SyncServer::DB

=item SyncServer::Config

=item SyncServer::LDAP

=back

=head1 ENVIRONMENT

=over 8

=item CONTENT_LENGTH

=item REQUEST_METHOD

=item SYNC_USER

=item AUTHORIZATION

=item REMOTE_PASSWD

=item REMOTE_USER

=back

=head1 METHODS

=over 8

=item B<new>()

Creates a new C<SyncServer> object, loads the configuration and prepares file/io handles. Also creates database and (where applicable) ldap objects. Hooks to C<FCGI>.

=item B<init>()

Cleans up the object, prepares to received data from C<FCGI> and refreshes the database connection.

=item B<header>(I<CODE>, I<MESSASGE>)

Sends the corresponding header to the sync-client.

=item B<send>(I<MESSAGE>, I<ARRAY>)

Sends I<MESSAGE> and I<ARRAY> to the sync-client.

=item  B<setcontentlength>(I<CONTENT_LENGTH>)

Sets the I<CONTENT_LENGTH> header.

=item B<setcontenttype>(I<CONTENT_TYPE>)

Sets the I<CONTENT_TYPE> header.

=item B<data>()

Reads data from the sync-client and decodes the JSON object if applicable.

=item B<dumpenv>()

For debugging purposes only, dumps the C<env> and C<qpars>.

=item B<env>(I<VAR>, I<VAL>)

Sets the environment.

=item B<qpars>(I<VAR>, I<VAL>)

Sets the qpars-environment.

=item B<requestIs>(I<REQUEST>)

Checks whether or not the request is a I<REQUEST>.

=item B<Accept>()

Wrapper for C<FCGI::Accept>.

=item B<Flush>()

Wrapper for C<FCGI::Flush>.

=item B<GetEnvironment>()

Wrapper for C<FCGI::GetEnvironment>.

=item B<auth>()

Sets environment variables and checks I<SYNC_USER>'s authentication.

=back

=head1 SEE ALSO

L<SyncServer::User>, L<SyncServer::Storage>, L<SyncServer::DB>, L<SyncServer::LDAP>, L<SyncServer::Config>, L<SyncServer::Err>

=head1 AUTHOR

Philippe Kueck <projects at unixadm dot org>

=cut
