# $Id: Config.pm 373 2013-07-02 13:37:14Z phil $
#
# author: Philippe Kueck <projects@unixadm.org>
#

package SyncServer::Config;

use strict;
use warnings;

our @ISA = qw/Exporter/;
our @EXPORT = qw/$config $dblock $emailrxp $collections $ldaptimeout/;

use constant CFGFILE => "/etc/syncserver/syncserver.cfg";

our $VERSION = "1.1";

our $config;
our $cfloaded :shared = 0;
our $dblock :shared = 0;
our $ldaptimeout :shared = 0;
our $emailrxp = qr/^[a-zA-Z0-9!#\$%&*+\/=?^_{|}~-]+(\.[a-zA-Z0-9!#\$%&*+\/=?^_{|}~-]+)?[@]([a-zA-Z0-9]([a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?\.)+[a-z]{1,63}$/;
our $collections = {
	1 => 'clients', 2 => 'crypto', 3 => 'forms', 4 => 'history', 5 => 'keys',
	6 => 'meta', 7 => 'bookmarks', 8 => 'prefs', 9 => 'tabs', 10 => 'passwords',
	11 => 'addons', 12 => 'adblockplus'
};


sub validate {
	return 1 unless $config->{$_[0]};
	my $ret = 1;
	foreach (@{$_[1]}) {
		unless ($config->{$_}) {
			printf STDERR "%s: %s missing\n", $_[0], $_;
			$ret = 0
		}
	}
	return $ret
}

sub chkmodule {
	return 1 unless $config->{$_[0]};
	unless (eval "require ".$_[1]) {
		printf STDERR "%s: %s wanted but not found\n", $_[0], $_[1];
		return 0
	}
	return 1
}

sub load {
	lock $cfloaded;
	return 1 if $cfloaded;
	my $bkeys = qr/^(?:register|recaptcha|ldap)$/;
	my $skeys = qr/^(?:secret|mailfrom|db(?:i|user|pass)|ldap(?:uri|bind(?:dn|pw)|basedn|filter)|recaptcha_p(?:riv|ub)key)$/;
	my $mkeys = qr/^(?:memcached)$/;
	my $true = qr/^(?:1|yes|true)$/i;
	my $line = 0;
	open CFG, "<" . CFGFILE or return 0;
	while (<CFG>) {
		++$line;
		next unless my ($k, $v) = /^\s*([a-z_-]+)\s*=\s*['"]?([^#'"]*?)['"]?(?:[\s#]+.*)?$/;
		if ($k =~ $bkeys) {$config->{$k} = ($v =~ $true)?1:0}
		elsif ($k =~ $skeys) {$config->{$k} = $v}
		elsif ($k =~ $mkeys) {push @{$config->{$k}}, $v}
		else {
			printf STDERR "parse error in config file line %d: %s\n",
			$line, $_
		}
	}
	close CFG;

	my $cfloaded_pre = 1;
	$cfloaded_pre &&= &validate('dbi', ['dbuser', 'dbpass']);
	$cfloaded_pre &&= &validate('ldap', ['ldapuri', 'ldapbasedn', 'ldapfilter', 'memcached']);
	$cfloaded_pre &&= &validate('recaptcha', ['recaptcha_privkey', 'recaptcha_pubkey']);
	$cfloaded_pre &&= &chkmodule('dbi', 'DBI');
	$cfloaded_pre &&= &chkmodule('ldap', 'Net::LDAP');
	$cfloaded_pre &&= &chkmodule('recaptcha', 'Captcha::reCAPTCHA');
	$cfloaded_pre &&= &chkmodule('memcached', 'Cache::Memcached');
	unless ($config->{'dbi'} && DBI->parse_dsn($config->{'dbi'})) {
		printf STDERR "dbi: cannot parse DBI\n"; $cfloaded_pre = 0
	}

	$config = undef unless $cfloaded_pre;
	$cfloaded = $cfloaded_pre;
	return $cfloaded
}

1;

__END__

=head1 NAME

SyncServer::Config - basic configuration, constants and semaphores

=head1 SYNOPSIS

 use SyncServer::Config;
 SyncServer::Config::load;

=head1 DESCRIPTION

This module is required for loading the configuration file, setting semaphores, and provides constants Syncserver uses elsewhere.

=head1 DEPENDENCIES

none.

=head1 ENVIRONMENT

none.

=head1 METHODS

=over 8

=item B<load>()

load and parses the configuration file (L</etc/syncserver/syncserver.cfg>).

=item B<validate>(I<STRING>, I<ARRAY>)

dependency check: if config option I<STRING> is set, check whether or not elements of I<ARRAY> are required and set.

=item B<chkmodule>(I<STRING>, I<MODULE>)

dependency check: if config option I<STRING> is set, check whether or not the perl I<MODULE> is available.

=back

=head1 AUTHOR

Philippe Kueck <projects at unixadm dot org>

=cut
