# $Id: DB.pm 386 2013-11-17 16:48:40Z phil $
#
# author: Philippe Kueck <projects@unixadm.org>
#

package SyncServer::DB;

use strict;
use warnings;

my $have_memcached = 0;
if (eval 'require Cache::Memcached') {
	use Cache::Memcached;
	$have_memcached = 1
}

use DBI;
use Digest::SHA qw(sha1);
use Time::HiRes qw(time);
use SyncServer::Config;


my $schema = {
	'users' => {
		'mysql' => qq{
			create table users (
			uid binary(20) not null primary key,
			username varchar(64) not null, password binary(20) not null,
			email varchar(255) not null, node varchar(255),
			resetcode char(16), resettime float
			) engine=innodb
		},
		'SQLite' => qq{
			create table users (
			uid binary(20) not null primary key,
			username varchar(64) not null, password binary(20) not null,
			email varchar(255) not null, node varchar(255),
			resetcode char(16), resettime float
			)
		}
	},
	'collections' => {
		'mysql' => qq{
			create table collections (
			id int unsigned not null auto_increment primary key,
			collection varchar(16) not null,
			unique key (collection)
			) engine=innodb
		},
		'SQLite' => qq{
			create table collections (
			id integer primary key,
			collection varchar(16) not null,
			unique (collection)
			)
		}
	},
	'wbo' => {
		'mysql' => qq{
			create table wbo (
		        uid binary(20) not null,
			collectiontype int unsigned not null,
		        wboid varchar(64) not null,
		        modified double not null,
		        sortindex int(9) not null,
		        payload mediumblob,
			payload_size int,
		        ttl int unsigned not null,
			parentid varchar(64),
			predecessorid varchar(64),
		        index (uid),
			unique index (uid, collectiontype, wboid),
		        foreign key (uid) references users(uid)
		                on delete cascade on update cascade,
			foreign key (collectiontype) references collections(id)
				on delete restrict on update cascade
			) engine=innodb
		},
		'SQLite' => qq{
			create table wbo (
		        uid binary(20) not null,
		        collectiontype int unsigned not null,
		        wboid varchar(64) not null,
		        modified double not null,
		        sortindex int(9) not null,
		        payload mediumblob,
			payload_size int,
		        ttl int unsigned not null,
			parentid varchar(64),
			predecessorid varchar(64),
		        foreign key (uid) references users(uid)
		                on delete cascade on update cascade,
			foreign key (collectiontype) references collections(id)
				on delete restrict on update cascade
			)
		}
	}
};


sub new {
	my $self = {
		'dbh' => DBI->connect_cached(
			$config->{'dbi'}, $config->{'dbuser'}, $config->{'dbpass'}
		)
	};
	$self->{'dbh'}->{'RaiseError'} = 1;
	$self->{'dbh'}->{'PrintError'} = 0;

	$self->{'cache'} = new Cache::Memcached({'servers' => $config->{'memcached'}})
		if $config->{'memcached'} && $have_memcached;

	return printf STDERR "Database error, please check settings!\n" unless $self->{'dbh'};

	lock($dblock);

	(undef, my $driver) = DBI->parse_dsn($config->{'dbi'});
	foreach my $table ('users', 'collections', 'wbo') {
		unless ($self->{'dbh'}->tables(undef, undef, $table, "TABLE")) {
			printf "table %s does not exist, creating\n", $table;
			$self->{'dbh'}->begin_work;
			my $sth = $self->{'dbh'}->prepare($schema->{$table}->{$driver});
			eval {$sth->execute};
			if ($@) {
				printf STDERR "Error creating table: %s\n", $@;
				$self->{'dbh'}->rollback;
				return
			}
			if ($table eq 'collections') {
				foreach my $i (keys %{$collections}) {
					$sth = $self->{'dbh'}->prepare_cached(qq{
						insert into collections values (?,?)
					});
					eval {$sth->execute($i, $collections->{$i})};
					if ($@) {
						printf STDERR "Error filling table: %s\n", $@;
						$self->{'dbh'}->rollback;
						return
					}
				}
			}
			$self->{'dbh'}->commit
			
		} 
	}
	return bless $self, $_[0]
}


sub reconnect {
	delete $_[0]->{'error'};
	$_[0]->{'dbh'} = DBI->connect_cached(
		$config->{'dbi'}, $config->{'dbuser'}, $config->{'dbpass'}
	);
	unless ($_[0]->{'dbh'}) {
		$_[0]->error(sprintf "(%d) %s", $DBI::err, $DBI::errstr);
		return 0
	}
	return 1
}

sub error {
	if ($_[1]) {
		printf STDERR "database error: %s\n", $_[1];
		$_[0]->{'error'} = $_[1]
	}
	return $_[0]->{'error'}
}


#######################################
# user functions
#

sub auth {
	lock($dblock);
        my $sth = $_[0]->{'dbh'}->prepare(qq{
		select uid from users where uid = ? and password = ?
	});
        eval {$sth->execute(sha1($_[1]), sha1($_[2]))};
	if ($@) {$_[0]->error($@); return 0}
	return $sth->fetchrow_array?1:0
}


sub getuser {
	my $obj;
	if ($config->{'memcached'}) {
		$obj = $_[0]->{'cache'}->get('sync_'.$_[1]);
		if ($obj && $obj->{'ts'} <= time - 20) {
			$_[0]->{'cache'}->delete('sync'.$_[1]); undef $obj
		}
		return $obj if defined $obj
	}

	lock($dblock);
        my $sth = $_[0]->{'dbh'}->prepare(qq{
		select uid, username, email, node from users where uid = ?
	});
        eval {$sth->execute(sha1($_[1]))};
	if ($@) {$_[0]->error($@); return}

        $obj = $sth->fetchrow_hashref();
        return unless $obj;
        $obj->{'ts'} = time;
        $_[0]->{'cache'}->set('sync_'.$_[1], $obj) if $config->{'memcached'};
        return $obj
}


sub setuser {
	lock($dblock);
	my $sth = $_[0]->{'dbh'}->prepare(qq{
		insert into users (uid, username, password, email, node) values (?,?,?,?,?)
	});
	$sth->bind_param(1, sha1($_[1]->{'username'}));
	$sth->bind_param(2, $_[1]->{'username'});
	$sth->bind_param(3, sha1($_[1]->{'password'}));
	$sth->bind_param(4, $_[1]->{'email'});
	$sth->bind_param(5, $_[1]->{'node'});
	unless (eval {$sth->execute}) {
		$_[0]->error($@) if $@;
		return 0
	}
	my $obj = {
		'uid' => sha1($_[1]->{'username'}),
		'username' => $_[1]->{'username'},
		'email' => $_[1]->{'email'},
		'node' => $_[1]->{'node'},
		'ts' => time
	};
	if ($config->{'memcached'}) {
		$_[0]->{'cache'}->delete('sync_'.$_[1]->{'username'});
		$_[0]->{'cache'}->set('sync_'.$_[1]->{'username'}, $obj)
	}
	return 1
}


sub setpass {
	lock($dblock);
	my $sth = $_[0]->{'dbh'}->prepare(qq{update users set password = ? where uid = ?});
	$sth->bind_param(1, sha1($_[2]));
	$sth->bind_param(2, sha1($_[1]));
	unless (eval {$sth->execute}) {
		$_[0]->error($@) if $@;
		return 0
	}
	return 1
}


sub setemail {
	lock($dblock);
	my $sth = $_[0]->{'dbh'}->prepare(qq{update users set email = ? where uid = ?});
	$sth->bind_param(1, $_[2]);
	$sth->bind_param(2, sha1($_[1]));
	unless (eval {$sth->execute}) {
		$_[0]->error($@) if $@;
		return 0
	}
	$_[0]->{'cache'}->delete('sync_'.$_[1]) if $config->{'memcached'};
	return 1
}


sub deluser {
	lock($dblock);
	my $sth = $_[0]->{'dbh'}->prepare(qq{delete from users where uid = ?});
	unless (eval {$sth->execute(sha1($_[1]))}) {
		$_[0]->error($@) if $@;
		return 0
	}
	$_[0]->{'cache'}->delete('sync_'.$_[1]) if $config->{'memcached'};
	return 1
}

sub getresetcode {
	lock($dblock);
        my $sth = $_[0]->{'dbh'}->prepare(qq{
		select resetcode, resettime from users where uid = ?
	});
        eval {$sth->execute(sha1($_[1]))};
	if ($@) {$_[0]->error($@); return}
        my $obj = $sth->fetchrow_hashref();
	return if !defined $obj || $obj->{'resetcode'} !~ /.+/ ||
		$obj->{'resetttime'} <= time - 86400;
	return $obj->{'resetcode'}
}

sub setresetcode {
	lock($dblock);
	my $sth = $_[0]->{'dbh'}->prepare(qq{
		update users set resetcode = ?, resettime = ? where uid = ?
	});
	$sth->bind_param(1, $_[2]);
	$sth->bind_param(2, time);
	$sth->bind_param(3, sha1($_[1]));
	unless (eval {$sth->execute}) {
		$_[0]->error($@) if $@;
		return 0
	}
	return 1
}


#######################################
# storage functions
#

sub infostor {
	my $ret;
	my $qry = {
		'collections' => qq{
			select collection, max(modified) as result from wbo
			left join collections on wbo.collectiontype = collections.id
			where uid = ? and ttl >= ? group by collection
		},
		'collection_usage' => qq{
			select collection, sum(payload_size)/1024 as result
			from wbo left join collections on wbo.collectiontype =
			collections.id where uid = ? and ttl >= ? group by collection
		},
		'collection_counts' => qq{
			select collection, count(wboid) as result from wbo
			left join collections on wbo.collectiontype = collections.id
			where uid = ? and ttl >= ? group by collection
		},
		'quota' => qq{
			select sum(length(payload))/1024 as result from wbo
			where uid = ? and ttl >= ?
		}
	};
	return unless exists $qry->{$_[2]};
	lock($dblock);
	my $sth = $_[0]->{'dbh'}->prepare($qry->{$_[2]});
        eval {$sth->execute(sha1($_[1]), time)};
	if ($@) {$_[0]->error($@); return}

	while (my $obj = $sth->fetchrow_hashref) {
		if ($_[2] eq 'quota') { # quota is kinda special..
			$ret = [$obj->{'result'}+0, undef];
			last
		}
		$ret->{$obj->{'collection'}} = $obj->{'result'}+0
	}
	return $ret
}

sub delstor {
	my $qry = {
		0 => qq{delete from wbo where uid = ? },
		1 => qq{delete from wbo where uid = ? and collectiontype =
		(select id from collections where collection = ?) },
		2 => qq{delete from wbo where uid = ? and wboid = ? and
		collectiontype = (select id from collections where collection = ?) },
		'ids' => qq{and wboid in (%s) }
	};
	return unless exists $qry->{scalar @{$_[2]}};
	my $crqry = $qry->{scalar @{$_[2]}};
	$crqry .= sprintf $qry->{'ids'}, join ',',
		map {$_ = "'$_'"} split ",", $_[3]->{'ids'}
		if $_[3]->{'ids'};
	lock($dblock);
	my $sth = $_[0]->{'dbh'}->prepare($crqry);
	$sth->bind_param(1, sha1($_[1]));
	$sth->bind_param(2, $_[2]->[0]) if defined $_[2]->[0];
	$sth->bind_param(3, $_[2]->[1]) if defined $_[2]->[1];
	
	eval {$sth->execute};
	if ($@) {
		$_[0]->error($@);
		return 0
	}
	return $sth->rows	
}

sub getstor {
	my ($ret, $obj);
	my $qry = {
		1 => qq{select DATA
			from wbo where uid = ? and collectiontype =
			(select id from collections where collection = ?)
			and ttl >= ?
		},
		2 => qq{select wboid as id, modified, payload
			from wbo where uid = ? and collectiontype =
			(select id from collections where collection = ?)
			and ttl >= ? and wboid = ?
		},
		'full#1' => qq{wboid as id, modified, payload},
		'full#0' => qq{wboid as id},
		'older' => qq{and modified < %d },
		'newer' => qq{and modified > %d },
		'index_below' => qq{and sortindex < %d },
		'index_above' => qq{and sortindex > %d },
		'ids' => qq{and wboid in (%s) },
		'limit' => qq{limit %d },
		'offset' => qq{offset %d },
		'sort#oldest' => qq{order by modified asc },
		'sort#newest' => qq{order by modified desc },
		'sort#index' => qq{order by sortindex desc }
	};
	return unless exists $qry->{scalar @{$_[2]}};
	my $crqry = $qry->{scalar @{$_[2]}};
	if ($_[3]->{'full'}) {$crqry =~ s/DATA/$qry->{'full#1'}/}
	else {$crqry =~ s/DATA/$qry->{'full#0'}/}
	foreach ('older', 'newer', 'index_below', 'index_above') {
		$crqry .= sprintf $qry->{$_}, $_[3]->{$_} if $_[3]->{$_}
	}

	$crqry .= sprintf $qry->{'ids'}, join ',',
		map {$_ = "'$_'"} split ",", $_[3]->{'ids'}
		if $_[3]->{'ids'};
	$crqry .= $qry->{'sort#'.$_[3]->{'sort'}} if $_[3]->{'sort'};
	$crqry .= sprintf $qry->{'limit'}, $_[3]->{'limit'} if $_[3]->{'limit'};
	$crqry .= sprintf $qry->{'offset'}, $_[3]->{'offset'} if $_[3]->{'offset'};
	lock($dblock);
	my $sth = $_[0]->{'dbh'}->prepare($crqry);
	$sth->bind_param(1, sha1($_[1]));
	$sth->bind_param(2, $_[2]->[0]);
	$sth->bind_param(3, time);
	$sth->bind_param(4, $_[2]->[1]) if defined $_[2]->[1];

	eval {$sth->execute};
	if ($@) {$_[0]->error($@); return}
	while ($obj = $sth->fetchrow_hashref) {
		if (scalar @{$_[2]} == 1 && !$_[3]->{'full'}) {
			push @{$ret}, $obj->{'id'}; next
		}
		$obj->{'modified'} += 0;
		push @{$ret}, $obj
	}

	return $ret->[0] if scalar @{$_[2]} == 2;
	return $ret
}

sub put_wbo {
	lock($dblock);
	# just do it, ignore errors
	(undef, my $driver) = DBI->parse_dsn($config->{'dbi'});
	if ($driver eq 'mysql') {
		$_[0]->{'dbh'}->do(
			sprintf "insert ignore into collections (collection) values ('%s')",
			$_[1]->{'collection'}
		)
	} elsif ($driver eq 'SQLite') {
		$_[0]->{'dbh'}->do(
			sprintf "insert or ignore into collections (collection) values ('%s')",
			$_[1]->{'collection'}
		)
	}
	my $dsth = $_[0]->{'dbh'}->prepare(qq{
		delete from wbo where uid = ? and wboid = ? and collectiontype =
		(select id from collections where collection = ?)
	});
	my $isth = $_[0]->{'dbh'}->prepare(qq{
		insert into wbo
		(uid, collectiontype, wboid, sortindex, payload, payload_size, modified, ttl)
		values (?, (select id from collections where collection = ?), ?, ?, ?, ?, ?, ?)
	});
	
	$dsth->bind_param(1, sha1($_[1]->{'user'}));
	$dsth->bind_param(2, $_[1]->{'id'});
	$dsth->bind_param(3, $_[1]->{'collection'});
	$isth->bind_param(1, sha1($_[1]->{'user'}));
	$isth->bind_param(2, $_[1]->{'collection'});
	$isth->bind_param(3, $_[1]->{'id'});
	$isth->bind_param(4, $_[1]->{'sortindex'});
	$isth->bind_param(5, $_[1]->{'payload'});
	$isth->bind_param(6, length $_[1]->{'payload'});
	$isth->bind_param(7, $_[1]->{'modified'});
	$isth->bind_param(8, $_[1]->{'ttl'} + $_[1]->{'modified'});

	eval {$dsth->execute; $isth->execute};

	$_[0]->error($@) if $@;
	return $isth->rows
}

sub update_wbo {
	lock($dblock);
	my $sth = $_[0]->{'dbh'}->prepare(qq{
		update wbo set sortindex = ?, modified = ?, ttl = ?
		where uid = ? wboid = ? and collectiontype =
		(select id from collections where collection = ?)
	});
	$sth->bind_param(1, $_[1]->{'sortindex'});
	$sth->bind_param(2, $_[1]->{'modified'});
	$sth->bind_param(3, $_[1]->{'ttl'} + $_[1]->{'modified'});
	$sth->bind_param(4, sha1($_[1]->{'user'}));
	$sth->bind_param(5, $_[1]->{'id'});
	$sth->bind_param(6, $_[1]->{'collection'});

	eval {$sth->execute};

	$_[0]->error($@) if $@;
	return $sth->rows
}

sub start_transaction {$_[0]->{'dbh'}->begin_work}
sub rollback_transaction {$_[0]->{'dbh'}->rollback}
sub commit_transaction {$_[0]->{'dbh'}->commit}

sub expire {
	lock($dblock);
	my $sth = $_[0]->{'dbh'}->prepare(qq{delete from wbo where ttl < ?});
	$sth->bind_param(1, time);
	
	eval {$sth->execute};

	$_[0]->error($@) if $@;
	return $sth->rows
}

1;

__END__

=head1 NAME

SyncServer::DB - SyncServer's DB interface

=head1 SYNOPSIS

 use SyncServer::DB;
 my $sdb = new SyncServer::DB;

=head1 DESCRIPTION

=head1 DEPENDENCIES

=over 8

=item Cache::Memcached

optional.

=item DBI

=item Digest::DHA

=item Time::HiRes

=item SyncServer::Config

=back

=head1 ENVIRONMENT

none

=head1 METHODS

=over 8

=item B<new>()

Creates a new C<SyncServer::DB> object and initializes the database.

=item B<reconnect>()

Refreshes the connection to the database.

=item B<error>(I<MESSAGE>)

Sets, prints and returns a database error.

=item B<auth>(I<USER>, I<PASSWORD>)

Authenticate user in database.

=item B<getuser>(I<USER>)

Get user data from DB or Memcached.

=item B<setuser>(I<HASH>)

Creates a new user. I<HASH> contains C<username>, C<password>, C<email> and C<node>.

=item B<setpass>(I<USER>, I<PASSWORD>)

Sets the I<USER>'s password.

=item B<setemail>(I<USER>, I<EMAIL>)

Sets the I<USER>'s email address.

=item B<deluser>(I<USER>)

Deletes I<USER>.

=item B<getresetcode>(I<USER>)

Returns I<USER>'s reset code.

=item B<setresetcode>(I<USER>, I<RESETCODE>)

Sets I<USER>'s reset code.

=item B<infostor>(I<TYPE>)

Returns information of I<TYPE>, where I<TYPE> is C<collections>, C<collection_usage>, C<collection_counts> or C<quota>.

=item B<delstor>(I<USER>, ...)

Deletes I<USER>'s storage objects. Additional parameters:

=over 8

=item none

Delete all WBOs

=item I<COLLECTION>

Delete all WBOs from I<COLLECTION>

=item I<COLLECTION>, I<WBOID>

Delete I<WBOID> from I<COLLECTION>

=back

=item B<getstor>(I<USER>, I<PARAMS>, I<HASH>)

Return records from database. I<PARAMS> can be:

=over 8

=item I<COLLECTION>

Return all from I<COLLECTION>.

=item I<COLLECTION>, I<WBOID>

Return I<WBOID> from I<COLLECTION>.

=back

I<HASH> can contain: 'full', 'older', 'newer', 'index_below', 'index_above', 'ids', 'limit' and 'offset'.

=item B<put_wbo>(I<HASH>)

Add WBO. I<HASH> contains: 'user', 'id', 'collection', 'sortindex', 'payload', 'modified' and 'ttl'.

=item B<update_wbo>(I<HASH>)

Update WBO. See B<put_wbo>().

=item B<start_transaction>()

Wrapper for C<DBI->begin_work>.

=item B<rollback_transaction>()

Wrapper for C<DBI->rollback>.

=item B<commit_transaction>()

Wrapper for C<DBI->commit_transaction>()

=item B<expire>()

Removes all WBOs with expired TTL.

=back

=head1 SEE ALSO

L<SyncServer>, L<SyncServer::Config>

=head1 AUTHOR

Philippe Kueck <projects at unixadm dot org>

=cut
