# $Id: LDAP.pm 384 2013-08-13 09:56:57Z phil $
#
# author: Philippe Kueck <projects@unixadm.org>
#

package SyncServer::LDAP;

use strict;
use warnings;

use Cache::Memcached;
use Net::LDAP;
use Digest::SHA qw(sha1);
use SyncServer::Config;


sub new {
	return bless {
		'cache' => new Cache::Memcached(
			{'servers' => $config->{'memcached'}}
		)
	}, $_[0]
}


sub refresh {
	lock($ldaptimeout);
	return unless time >= $ldaptimeout + 30;
	$ldaptimeout = time;
	return unless $_[0]->connect;
	my ($ret, $username);

	$ret = $_[0]->{'ldap'}->search(
		'base' => $config->{'ldapbasedn'},
		'filter' => $config->{'ldapfilter'},
		'attrs' => ['mail']
	);

	if ($ret->code) {
		$_[0]->error(sprintf "(%d) %s", $ret->code, $ret->error);
		return 0
	}

	$ret = $ret->as_struct;
	foreach (keys %{$ret}) {
		$username = unpack 'B*', sha1(lc $ret->{$_}->{'mail'}[0]);
		$username =~ s/(.....)/000$1/g;
		$username = pack 'B*', $username;
		$username =~ tr|\0-\37|A-Z2-7|;

		$_[0]->{'cache'}->set('lsync_'. lc $username, {
			'username' => lc $username,
			'email' => $ret->{$_}->{'mail'}[0],
			'dn' => $_,
			'ts' => 2147483647
		})
	}

	$_[0]->disconnect
}

sub connect {
	delete $_[0]->{'error'};

	$_[0]->{'ldap'} = new Net::LDAP($config->{'ldapuri'});
	unless ($_[0]->{'ldap'}) {$_[0]->error($@); return 0}

	my $ret;
	if ($config->{'ldapbinddn'} && $config->{'ldapbindpw'}) {
		$ret = $_[0]->{'ldap'}->bind(
			$config->{'ldapbinddn'},
			'password' => $config->{'ldapbindpw'}
		)
	} else {
		$ret = $_[0]->{'ldap'}->bind
	}
	if ($ret->code) {
		$_[0]->error(sprintf "(%d) %s", $ret->code, $ret->error);
		return 0
	}

	return 1
}

sub disconnect {
	$_[0]->{'ldap'}->unbind
}

sub error {
	if ($_[1]) {
		printf STDERR "LDAP error: %s\n", $_[1];
		$_[0]->{'error'} = $_[1]
	}
	return $_[0]->{'error'}
}

sub auth {
	my $ldap = new Net::LDAP($config->{'ldapuri'});
	unless ($ldap) {$_[0]->error($@); return 0}
	my $obj = $_[0]->{'cache'}->get('lsync_'.$_[1]);
	return 0 unless $obj;
	my $ret = $ldap->bind($obj->{'dn'}, 'password' => $_[2]);
	$ldap->unbind;
	return 0 if $ret->code;
	return 1
}

sub getuser {
	for (1..2) {
		my $obj = $_[0]->{'cache'}->get('lsync_'.$_[1]);
		return $obj if $obj;
		$_[0]->refresh
	}
}

sub setuser {
	return
}

sub setpass {
	return
}

sub setemail {
	return
}

sub deluser {
	$_[0]->{'cache'}->delete('lsync_'.$_[1]);
	return 1
}

sub getresetcode {
	return
}

sub setresetcode {
	return
}


1;

__END__

=head1 NAME

SyncServer::LDAP - The LDAP interface for user management

=head1 SYNOPSIS

 use SyncServer::LDAP;
 my $handler = new SyncServer::LDAP;

=head1 DESCRIPTION

This module provides the ldap interface. It is required when using an ldap database instead of a mysql/sqlite.

=head1 DEPENDENCIES

=over8

=item Net::LDAP

=item Cache::Memcached

=item Digest::SHA

=item SyncServer::Config

=back

=head1 ENVIRONMENT

none

=head1 METHODS

=over 8

=item B<new>()

Creates a new ldap object which contains a L<Cache::Memcached> object.

=item B<refresh>()

Refreshes the ldap contents to the L<Cache::Memcached> object.

=item B<connect>()

Connects and binds to the given ldap server.

=item B<disconnect>()

Disconnects from the given ldap server.

=item B<error>(I<MESSAGE>)

Save I<MESSAGE> to the object and prints it to C<STDOUT>.

=item B<auth>(I<USER>, I<PASSWORD>)

Binds to the ldap server with the given credentials in order to check the authentication.

=item B<getuser>(I<USER>)

Returns the I<USER> record if it exists. Else refresh and try again.

=item B<setuser>()

Not implemented yet.

=item B<setpass>()

Not implemented yet.

=item B<setemail>()

Not implemented yet.

=item B<deluser>(I<USER>)

Delete I<USER> from Memcache.

=item B<getresetcode>()

Not implemented yet.

=item B<setresetcode>()

Not implemented yet.

=back

=head1 SEE ALSO

L<SyncServer>, L<SyncServer::Config>

=head1 AUTHOR

Philippe Kueck <projects at unixadm dot org>

=cut
