# $Id: Storage.pm 367 2013-07-02 13:12:16Z phil $
#
# author: Philippe Kueck <projects@unixadm.org>
#

package SyncServer::Storage;

use strict;
use warnings;
use JSON;
use Time::HiRes qw(time);
use SyncServer;
use SyncServer::Config;
use SyncServer::Err;

our @ISA = qw(SyncServer);


sub new {
	my $class = shift;
	my $super = $class->SUPER::new(@_);
	return $super
}

# XXX quick'n'dirty hack, erstmal collections beschraenken
sub validcollection {
	foreach my $c (keys %{$SyncServer::Config::collections}) {
		return 1 if $SyncServer::Config::collections->{$c} eq $_[1]
	}
	return 0
}

sub info {
	my $obj = $_[0]->{'db'}->infostor($_[0]->{'env'}->{'SYNC_USER'}, $_[1]);
	unless ($obj) {
		return new SyncServer::Err(503) if $_[0]->{'db'}->error;
		$obj = {}
	}
	$_[0]->{'content-type'} = "application/json";
	return (1, encode_json($obj))
}

sub chkprecond {
	return 1 unless $_[0]->{'env'}->{'HTTP_X_IF_UNMODIFIED_SINCE'};
	my $obj = $_[0]->{'db'}->infostor(
		$_[0]->{'env'}->{'SYNC_USER'}, "collections"
	);
	return 1 unless exists $obj->{$_[0]->{'env'}->{'COLLECTION'}};
	return $obj->{$_[0]->{'env'}->{'COLLECTION'}} >
		$_[0]->{'env'}->{'HTTP_X_IF_UNMODIFIED_SINCE'}
}

sub delete {
	my $self = shift;
	return new SyncServer::Err(412, "Precondition Failed")
		unless $self->chkprecond;
	my $ret = $self->{'db'}->delstor(
		$self->{'env'}->{'SYNC_USER'}, [@_], $self->{'qpars'}
	);
	return new SyncServer::Err(503) if !$ret && $self->{'db'}->error;
	return ($ret, sprintf("%.2f", time)+0)
}

sub put_collection {
	return new SyncServer::Err(412, "Precondition Failed")
		unless $_[0]->chkprecond;
	my ($ret, $obj);
	$obj->{'modified'} = sprintf("%.2f", time)+0;
	$obj->{'failed'} = {};
	$obj->{'success'} = [];
	return new SyncServer::Err(400, "6 JSON parse failure") unless
		ref $_[0]->{'data'} eq 'ARRAY';
	$_[0]->{'db'}->start_transaction;
	foreach my $wbo (@{$_[0]->{'data'}}) {
		$wbo->{'modified'} = $obj->{'modified'};
		($ret) = $_[0]->put_wbo($wbo);
		if ($ret) {push @{$obj->{'success'}}, $wbo->{'id'}}
		else {
			return new SyncServer::Err(503) if $_[0]->{'db'}->error;
			$obj->{'failed'}->{$wbo->{'id'}} = ['bogus']
		}
	}
	$_[0]->{'db'}->commit_transaction;

	$_[0]->{'content-type'} = "application/json";
	return (1, encode_json($obj))
}

sub put_wbo {
	my ($wbo, $ts) = (undef, sprintf("%.2f", time)+0);
	if (ref $_[1] && ref $_[1] eq 'HASH') {
		$wbo = $_[1]
	} elsif (ref $_[0]->{'data'} && ref $_[0]->{'data'} eq 'HASH') {
		return new SyncServer::Err(412, "Precondition Failed")
			unless $_[0]->chkprecond;
		$wbo = $_[0]->{'data'}
	} else {
		return new SyncServer::Err(400, "8 Invalid Weave Basic Object")
	}
	$wbo->{'user'} = $_[0]->{'env'}->{'SYNC_USER'};
	$wbo->{'collection'} = $_[0]->{'env'}->{'COLLECTION'};
	$wbo->{'id'} ||= $_[0]->{'env'}->{'WBOID'};
	$wbo->{'modified'} ||= $ts;
	$wbo->{'sortindex'} ||= 0;
	$wbo->{'ttl'} ||= 16380000;
	return (0, $ts) unless $wbo->{'collection'} && $wbo->{'id'};
	my $ret;
	if ($wbo->{'payload'}) {$ret = $_[0]->{'db'}->put_wbo($wbo)}
	else {$ret = $_[0]->{'db'}->update_wbo($wbo)}
	return new SyncServer::Err(503) if !$ret && $_[0]->{'db'}->error;
	return ($ret, $ts)
}

sub get {
	my $self = shift;
	my $obj = $self->{'db'}->getstor(
		$self->{'env'}->{'SYNC_USER'}, [@_], $self->{'qpars'}
	);
	unless ($obj) {
		return new SyncServer::Err(503) if $self->{'db'}->error;
		# even though the api allows us to return a 404 for empty
		# collections, firefox throws an error, refusing to sync.
		# happened here: tabs ttl'ed out after not using the
		# browser for a week, get ... tabs/full=1 returned a 404
		# which firefox didn't like.
		# return 1
		# XXX as of firefox 16/17/whatever it works as expected
		return 0
	}
	if (ref $obj eq 'ARRAY') {
		$self->{'content-type'} = "application/newlines";
		my $ret;
		foreach (@{$obj}) {$ret .= encode_json($_)."\n"}
		return (1, $ret)
	}
	if (ref $obj eq 'HASH') {
		$self->{'content-type'} = "application/json";
		return (1, encode_json($obj))
	}
	return new SyncServer::Err(503)
}

1;

__END__

=head1 NAME

SyncServer::Storage - Core module for storage

=head1 SYNOPSIS

 use SyncServer::Storage;
 my $syncserverid = new SyncServer::Storage;

=head1 DESCRIPTION

This module provides interfaces for storage management, commonly used by the storserver.

=head1 DEPENDENCIES

=over 8

=item JSON

=item Time::HiRes

=item SyncServer

=item SyncServer::Config

=item SyncServer:Err

=back

=head1 ENVIRONMENT

=over 8

=item SYNC_USER

=item HTTP_X_IF_UNMODIFIED_SINCE

=item COLLECTION

=item WBOID

=back

=head1 METHODS

=over 8

=item B<new>()

Creates a new C<SyncServer::Storage> object.

=item B<validcollection>()

Checks whether or not a valid collection was provided (obsoleted).

=item B<info>()

Gets information about the I<SYNC_USER>'s storage.

=item B<chkprecond>()

Checks whether or not all needed information is provided.

=item B<delete>()

Delete WBOs / collection from storage.

=item B<put_collection>()

Put collection to storage (calls B<put_wbo>()).

=item B<put_wbo>(I<HASH>), B<put_wbo>(I<ARRAY>)

Put WBOs to storage.

=item B<get>()

Get WBOs from storage.

=back

=head1 SEE ALSO

L<SyncServer>, L<SyncServer::Config>, L<SyncServer:Err>

=head1 AUTHOR

Philippe Kueck <projects at unixadm dot org>

=cut
