# $Id: User.pm 367 2013-07-02 13:12:16Z phil $
#
# author: Philippe Kueck <projects@unixadm.org>
#

package SyncServer::User;

use strict;
use warnings;
use Mail::Sendmail;
my $have_recaptcha = 0;
if (eval 'require Captcha::reCAPTCHA') {
	use Captcha::reCAPTCHA;
	$have_recaptcha = 1;
}
use SyncServer;
use SyncServer::Config;
use SyncServer::Err;

our @ISA = qw(SyncServer);


sub new {
	my $class = shift;
	my $super = $class->SUPER::new(@_);
	return $super
}


sub check_captcha {
	return 1 unless $config->{'recaptcha'} && $have_recaptcha;
	return 1 if $config->{'secret'} && $_[0]->{'env'}->{'HTTP_X_WEAVE_SECRET'} &&
		$_[0]->{'env'}->{'HTTP_X_WEAVE_SECRET'} eq $config->{'secret'};
	return 0 unless $_[0]->{'data'}->{'captcha-challenge'} &&
		$_[0]->{'data'}->{'captcha-response'};

	my $captcha = new Captcha::reCAPTCHA;
	my $ret = $captcha->check_answer(
		$config->{'recaptcha_privkey'}, $_[0]->{'env'}->{'REMOTE_ADDR'},
		$_[0]->{'data'}->{'captcha-challenge'},
		$_[0]->{'data'}->{'captcha-response'}
	);
	return $ret->{'is_valid'}?1:0
}

sub check_reset_code {
	return 0 if $config->{'ldap'};
	return 0 unless $_[0]->{'env'}->{'HTTP_X_WEAVE_PASSWORD_RESET'};
	return ($_[0]->{'env'}->{'HTTP_X_WEAVE_PASSWORD_RESET'} eq
		$_[0]->{'db'}->getresetcode($_[0]->{'env'}->{'SYNC_USER'}))?1:0
}

sub getpwquality {
	return 0 if length $_[1] < 8;
	return 1;
	my ($num, $upper, $special) = (1, 1, 1);
	foreach (split //, $_[1]) {
		if (/^\d$/) {$num--; next}
		if (/^[+\/!§\$%&()=?*'#,.;:<>{}~|_-]$/) {$special--; next}
		if (uc $_ eq $_) {$upper--; next}
	}
	return 1 if ($num <= 0 && $upper <= 0 && $special <= 0);
	return 0
}

sub exists {
	my $ret = $_[0]->{'db'}->getuser($_[0]->{'env'}->{'SYNC_USER'});
	if (!$ret && $_[0]->{'db'}->error) {
		printf STDERR "database error: %s\n", $_[0]->{'db'}->error;
		return new SyncServer::Err(503)
	}
	return $ret?1:0
}

sub registered {
	return (1, $_[0]->exists)
}

sub create {
	return new SyncServer::Err(503, "LDAP-User does not exist")
		if $config->{'ldap'} &&
		!$_[0]->{'ldap'}->getuser($_[0]->{'env'}->{'SYNC_USER'});
	return new SyncServer::Err(503, "Registration closed") unless
		$config->{'register'};
	return new SyncServer::Err(400, "6 JSON parse failure")
		unless $_[0]->{'data'};
	return new SyncServer::Err(400, "7 Missing password field")
		unless $_[0]->{'data'}->{'password'};
	return new SyncServer::Err(400, "12 No email address on file")
		unless $_[0]->{'data'}->{'email'} &&
		$_[0]->{'data'}->{'email'} =~ $emailrxp;
	# when using ldap the password is ignored
	return new SyncServer::Err(400, "9 Requested password not strong enough")
		unless $config->{'ldap'} ||
		$_[0]->getpwquality($_[0]->{'data'}->{'password'});
	return new SyncServer::Err(400, "2 Incorrect or missing captcha")
		unless $_[0]->check_captcha;
	my $ret = $_[0]->{'db'}->setuser({
		'username' => $_[0]->{'env'}->{'SYNC_USER'},
		'password' => $config->{'ldap'}?"":$_[0]->{'data'}->{'password'},
		'email' => $_[0]->{'data'}->{'email'},
		'node' => $_[0]->{'env'}->{'HTTP_HOST'}
	});
	unless ($ret) {
		return new SyncServer::Err(503) if $_[0]->{'db'}->error;
		return new SyncServer::Err(400, "4 User already exists")
	}
	return (1, lc $_[0]->{'env'}->{'SYNC_USER'})
}

sub delete {
	return new SyncServer::Err(503, "Registration closed") unless
		$config->{'register'};
	my $ret = $_[0]->{'db'}->deluser($_[1]);
	$ret &&= $_[0]->{'ldap'}->deluser($_[1]) if $config->{'ldap'};
	unless ($ret) {
		return new SyncServer::Err(503) if $_[0]->{'db'}->error;
		return new SyncServer::Err(404, "The user does not exist")
	}
	return (1, 0)
}

sub getnode {
	my $ret = $_[0]->{'db'}->getuser($_[0]->{'env'}->{'SYNC_USER'});
	unless ($ret) {
		return new SyncServer::Err(503) if $_[0]->{'db'}->error;
		return new SyncServer::Err(404, "The user does not exist")
	}
	return (1, sprintf "https://%s/storage/",
		$ret->{'node'}?$ret->{'node'}:$_[0]->{'env'}->{'HTTP_HOST'})
}

sub setpassword {
	return new SyncServer::Err(503, "Backend does not support this yet")
		if $config->{'ldap'};
	return new SyncServer::Err(400, "10 Invalid or missing password reset code")
		unless $_[0]->{'env'}->{'AUTHORIZATION'} || $_[0]->check_reset_code;
	return new SyncServer::Err(400, "7 Missing password field")
		unless $_[0]->{'data'};
	return new SyncServer::Err(400, "9 Requested password is not strong enough")
		unless $_[0]->getpwquality($_[0]->{'data'});
	my $ret = $_[0]->{'db'}->setpass($_[0]->{'env'}->{'SYNC_USER'}, $_[0]->{'data'});
	unless ($ret) {
		return new SyncServer::Err(503) if $_[0]->{'db'}->error;
		return new SyncServer::Err(503, "There was an error updating the password")
	}
	return (1, "success")
}

sub setemail {
	return new SyncServer::Err(503, "Backend does not support this yet")
		if $config->{'ldap'};
	return new SyncServer::Err(400, "12 No email address on file")
		unless $_[0]->{'data'} && $_[0]->{'data'} =~ $emailrxp;
	my $ret = $_[0]->{'db'}->setemail($_[0]->{'SYNC_USER'}, $_[0]->{'data'});
	unless ($ret) {
		return new SyncServer::Err(503) if $_[0]->{'db'}->error;
		return new SyncServer::Err(503, "There was an error updating the email address")
	}
	return (1, $_[0]->{'data'})
}

sub send_password_reset {
	return new SyncServer::Err(503, "Backend does not support this yet")
		if $config->{'ldap'};
	my $data = $_[0]->{'db'}->getuser($_[0]->{'env'}->{'SYNC_USER'});
	return new SyncServer::Err(400, "2 Incorrect or missing captcha") unless
		$_[0]->check_captcha;
	return new SyncServer::Err(400, "12 No email address on file")
		unless $data->{'email'};
	my $code;
	my @chars = ('A'..'Z', 'a'..'z', '0'..'9', '_', '-', '/', '+', '%', '$', '!');
	foreach (0..16) {$code .= $chars[rand(@chars)]}
	my $ret = $_[0]->{'db'}->setresetcode($_[0]->{'env'}->{'SYNC_USER'}, $code);
	unless ($ret) {
		return new SyncServer::Err(503) if $_[0]->{'db'}->error;
		return new SyncServer::Err(503, "Database update failed")
	}
	
	my %mail = (
		'X-Mailer' => 'SyncServer',
		'To' => $data->{'email'},
		'From' => $config->{'mailfrom'},
		'Subject' => 'Your password reset code',
		'Message' => sprintf "Hello, this is your SyncServer.\n".
			"You have requested a password reset. This code is\n".
			"\n%s\n\nKind regards,\nSyncServer", " "x15 . $code
	);
	return new SyncServer::Err() unless sendmail %mail;
	return 1
}


1;

__END__

=head1 NAME

SyncServer::User - Core module for user management

=head1 SYNOPSIS

 use SyncServer::User;
 my $syncserverid = new SyncServer::User;

=head1 DESCRIPTION

This module provides interfaces for user management, commonly used by the regserver.

=head1 DEPENDENCIES

=over 8

=item SyncServer

=item SyncServer::Config

=item SyncServer::Err

=item Mail::Sendmail

=item Captcha::reCAPTCHA

(optional)

=back

=head1 ENVIRONMENT

=over 8

=item HTTP_X_WEAVE_SECRET

Password for bypassing the reCAPTCHA.

=item REMOTE_ADDR

Peer address.

=item HTTP_X_WEAVE_PASSWORD_RESET

Flag for resetting the password.

=item SYNC_USER

The syncserver user.

=item HTTP_HOST

That's me.

=item AUTHORIZATION

The base64 encoded authorization header.

=back

=head1 METHODS

=over 8

=item B<new>()

Create a new C<SyncServer::User> object.

=item B<check_captcha>()

Checks whether or not the given I<captcha-challenge> matches the I<captcha-response>. See Captcha::reCAPTCHA for details.

=item B<check_reset_code>()

Compares the provided I<HTTP_X_WEAVE_PASSWORD_RESET> with the stored one.

=item B<getpwquality>(I<PASSWORD>)

Checks the quality of the given I<PASSWORD>: At least one number, one upper case character and one special character is required.

=item B<exists>()

Checks whether or not the I<SYNC_USER> exists.

=item B<registered>()

Wrapper for B<exists>().

=item B<create>()

Creates the given I<SYNC_USER>.

=item B<delete>(I<USER>)

Deletes the given I<USER>.

=item B<getnode>()

Returns the storage node the I<SYNC_USER> resides on.

=item B<setpassword>()

Sets the I<SYNC_USER>'s password.

=item B<setemail()>

Sets the I<SYNC_USER>'s email address.

=item B<send_password_reset>()

Sends the password reset code to I<SYNC_USER>'s email address.

=back

=head1 SEE ALSO

L<SyncServer>, L<SyncServer::Config>, L<SyncServer::Err>

=head1 AUTHOR

Philippe Kueck <projects at unixadm dot org>

=cut
